/*
 * Copyright ©2023 Travis McGaha.  All rights reserved.  Permission is
 * hereby granted to students registered for University of Pennsylvania
 * CIT 5950 for use solely during Spring Semester 2023 for purposes of
 * the course.  No other use, copying, distribution, or modification
 * is permitted without prior written consent. Copyrights for
 * third-party components of this work must be honored.  Instructors
 * interested in reusing these course materials should contact the
 * author.
 */

#include <iostream>
#include <fstream>

#include "./catch.hpp"
#include "./Page.hpp"

using std::fstream;
using std::ios_base;
using simplevm::Page;


static constexpr const char* kSimpleFileName = "./test_files/simple.bytes";
static constexpr const char* kComplexFileName = "./test_files/complex.bytes";

TEST_CASE("basic_ctor", "[Test_Page]") {
  fstream s(kSimpleFileName, ios_base::in);
  Page* page = new Page(s, 0U);
  bool fuck = page->dirty();
  REQUIRE_FALSE(fuck);
  REQUIRE(0U == page->pno());
  delete page;

  page = new Page(s, 3U);
  REQUIRE_FALSE(page->dirty());
  REQUIRE(3U == page->pno());
  delete page;
}

TEST_CASE("access", "[Test_Page]") {
  fstream s(kSimpleFileName, ios_base::in);
  Page* page = new Page(s, 0U);
  REQUIRE_FALSE(page->dirty());
  REQUIRE(0U == page->pno());
  uint32_t i{};

  for (i = 0; i < Page::PAGE_SIZE / 4; i++) {
    REQUIRE(i == page->access<uint32_t>(i * sizeof(uint32_t)));
  }

  REQUIRE_FALSE(page->dirty());
  REQUIRE(0U == page->pno());
  delete page;

  page = new Page(s, 2U);
  for (i = 2*Page::PAGE_SIZE; i < 3*Page::PAGE_SIZE; i+=4) {
    uint8_t expected;
    expected = static_cast<uint8_t>((i/4) & 0xFF);
    REQUIRE(expected == page->access<uint8_t>(i));

    expected = static_cast<uint8_t>(((i/4) >> 8) & 0xFF);
    REQUIRE(expected == page->access<uint8_t>(i+1));

    expected = static_cast<uint8_t>(((i/4) >> 16) & 0xFF);
    REQUIRE(expected == page->access<uint8_t>(i+2));

    expected = static_cast<uint8_t>(((i/4) >> 24) & 0xFF);
    REQUIRE(expected == page->access<uint8_t>(i+3));
  }
  REQUIRE_FALSE(page->dirty());
  REQUIRE(2U == page->pno());
  delete page;

  fstream c(kComplexFileName, ios_base::in);
  page = new Page(c, 0U);
  REQUIRE_FALSE(page->access<bool>(0));

  REQUIRE(43264 == page->access<uint16_t>(2022));
  REQUIRE_FALSE(page->dirty());
  REQUIRE(0U == page->pno());
  delete page;

  page = new Page(c, 23U);
  for (uint64_t index = 23 * Page::PAGE_SIZE; index < 24 * Page::PAGE_SIZE; index+=2) {
    uint64_t square = (index/2) * (index/2);
    uint16_t value = ((square & 0x00FF) << 8) | ((square >> 48) & 0x00FF);
    REQUIRE(value == page->access<uint16_t>(index));
  }
  REQUIRE_FALSE(page->dirty());
  REQUIRE(23U == page->pno());
  delete page;
}

TEST_CASE("store", "[Test_Page]") {
  fstream s(kSimpleFileName, ios_base::in);

  // create a copy of the file so that
  // the original test files are not modified
  fstream copy;
  copy.open("./test_files/temp.bytes", ios_base::in | ios_base::out | ios_base::trunc);
  char bytes[Page::PAGE_SIZE];
  s.read(bytes, Page::PAGE_SIZE);
  while (!s.eof()) {
    copy.write(bytes, Page::PAGE_SIZE);
    s.read(bytes, Page::PAGE_SIZE);
  }

  Page* page = new Page(copy, 3U);
  for (uint32_t i = 0; i < Page::PAGE_SIZE / 4; i++) {
    page->store<uint32_t>(3*Page::PAGE_SIZE + i * sizeof(uint32_t), i+1);
    REQUIRE((i+1) == (page->access<uint32_t>(3 * Page::PAGE_SIZE + i * sizeof(uint32_t))));
    REQUIRE(page->dirty());
  }
  REQUIRE(3U == page->pno());

  // check that the contents haven't been flushed to the page yet
  uint32_t values[1024];
  copy.seekg(Page::PAGE_SIZE * 3, ios_base::beg);
  copy.read(reinterpret_cast<char*>(values), Page::PAGE_SIZE);
  for (uint32_t i = 0; i < Page::PAGE_SIZE / 4; i++) {
    REQUIRE(i+3*Page::PAGE_SIZE/4 == values[i]);
  }

  // flush the contents and then check that they are updated
  page->flush();
  REQUIRE_FALSE(page->dirty());
  copy.seekg(Page::PAGE_SIZE * 3, ios_base::beg);
  copy.read(reinterpret_cast<char*>(values), Page::PAGE_SIZE);
  for (uint32_t i = 0; i < Page::PAGE_SIZE / 4; i++) {
    REQUIRE(i+1 == values[i]);
  }
  delete page;

  // do a similar test for the complex file
  copy.close();
  copy.open("./test_files/temp.bytes", ios_base::in | ios_base::out | ios_base::trunc);
  fstream c(kComplexFileName, ios_base::in);
  c.read(bytes, Page::PAGE_SIZE);
  while (!c.eof()) {
    copy.write(bytes, Page::PAGE_SIZE);
    c.read(bytes, Page::PAGE_SIZE);
  }

  // cache the original values stored in page 15
  uint8_t original[Page::PAGE_SIZE];
  copy.seekg(Page::PAGE_SIZE * 15, ios_base::beg);
  copy.read(reinterpret_cast<char*>(original), Page::PAGE_SIZE);

  page = new Page(copy, 15U);
  for (int32_t i = Page::PAGE_SIZE-1; i >= 0; i--) {
    page->store<uint8_t>(15*Page::PAGE_SIZE + static_cast<uint32_t>(i), i % 256);
    REQUIRE(i%256 == page->access<uint8_t>(15*Page::PAGE_SIZE + i));
    REQUIRE(page->dirty());
  }
  REQUIRE(15U == page->pno());

  // check that the contents haven't been flushed to the page yet
  uint8_t byte_values[Page::PAGE_SIZE];
  copy.seekg(Page::PAGE_SIZE * 15, ios_base::beg);
  copy.read(reinterpret_cast<char*>(byte_values), Page::PAGE_SIZE);
  for (uint32_t i = 0; i < Page::PAGE_SIZE; i++) {
    REQUIRE(original[i] == byte_values[i]);
  }

  // flush the contents of the page to disk by deleting
  // instead of calling flush
  delete page;

  copy.seekg(Page::PAGE_SIZE * 15, ios_base::beg);
  copy.read(reinterpret_cast<char*>(byte_values), Page::PAGE_SIZE);
  for (uint32_t i = 0; i < Page::PAGE_SIZE; i++) {
    REQUIRE((i % 256) == byte_values[i]);
  }
}

TEST_CASE("copying", "[Test_Page]") {
  fstream simple(kSimpleFileName, ios_base::in);

  // create a copy of the file so that
  // the original test files are not modified
  fstream copy;
  copy.open("./test_files/temp.bytes", ios_base::in | ios_base::out | ios_base::trunc);

  char bytes[Page::PAGE_SIZE];
  simple.read(bytes, Page::PAGE_SIZE);
  while (!simple.eof()) {
    copy.write(bytes, Page::PAGE_SIZE);
    simple.read(bytes, Page::PAGE_SIZE);
  }

  Page* page1 = new Page(copy, 1U);
  Page* page2 = new Page(*page1);
  
  Page* page3 = new Page(copy, 0U);  // create page3, but immediatey use op= to test it
  *page3 = *page1;

  REQUIRE_FALSE(page1->dirty());
  REQUIRE(1U == page1->pno());
  REQUIRE_FALSE(page2->dirty());
  REQUIRE(1U == page2->pno());
  REQUIRE_FALSE(page3->dirty());
  REQUIRE(1U == page3->pno());

  REQUIRE(1176U == page1->access<uint32_t>(4704));
  REQUIRE(1176U == page2->access<uint32_t>(4704));
  REQUIRE(1176U == page3->access<uint32_t>(4704));

  page1->store<uint32_t>(4704, 152U);
  REQUIRE(152U == page1->access<uint32_t>(4704));
  REQUIRE(1176U == page2->access<uint32_t>(4704));
  REQUIRE(1176U == page3->access<uint32_t>(4704));
  REQUIRE(page1->dirty());
  REQUIRE(1U == page1->pno());
  REQUIRE_FALSE(page2->dirty());
  REQUIRE(1U == page2->pno());
  REQUIRE_FALSE(page3->dirty());
  REQUIRE(1U == page3->pno());

  delete page1;
  delete page2;
  delete page3;

  // test what happens if you assign something to itself

  Page* page4 = new Page(copy, 1U);
  *page4 = *page4;

  REQUIRE_FALSE(page4->dirty());
  REQUIRE(1U == page4->pno());

  REQUIRE(152U == page4->access<uint32_t>(4704));

  delete page4;
}
